%ESN_EFFICIENCYCURVE    Given a filename of results of a certain format,
%this script loads the data and plots an efficiency curve for the random
%search.
%
% The following variables are set by the user:
% path2DB   - Path containing the ESN data/results
% fn        - String (or cell array of strings) containing the name of
%           result files. The result files should contain:
%               Se              - Nx1 vector of DB1 Sensitivities
%               PPV             - Nx1 vector of DB1 Positive predictive val
%               F1Score         - Nx1 vector of DB1 F1 measures
%               Se_DB2          - Nx1 vector of DB2 Sensitivities
%               PPV_DB2         - Nx1 vector of DB2 Positive predictive val
%               F1Score_DB2     - Nx1 vector of DB2 F1 measures
%               stats_db1       - NxD structure array with statistics
%                               calculated on each individual record in DB1
%               stats_db2       - NxM structure array with statistics
%                               calculated on each individual record in DB2
%               params          - Structure containing Nx1 vectors of each
%                               hyperparameter that changed during the search
%       ... where N is the number of random searches performed

%	Copyright 2013 Alistair Johnson

%	$LastChangedBy: alistair $
%	$LastChangedDate: 2013-11-28 14:24:36 +0000 (Thu, 28 Nov 2013) $
%	$Revision: 960 $
%	Originally written on GLNXA64 by Alistair Johnson, 16-Oct-2013 16:01:31
%	Contact: alistairewj@gmail.com

%%  --------------- HEADER ---------------
clearvars -except path2DB fn; close all; clc;

if exist('fn','var')~=1 % Default file name
fn = 'SummaryResults.mat'; % repetition of 512 runs
end

if exist('path2DB','var')~=1 % Default path
path2DB = ''; % absolute path to database
end


% Set the paths containing the ESN database
DOWN_SAMPLE = 1;                 % run all the algo at 250Hz or 1000Hz
N_DB1 = 14; % records in database 1
N_DB2 = 11; % records in database 2

addpath(genpath(path2DB)); 

%%  --------------- LOAD RESULTS ---------------  
%=== Ensure the file name(s) are valid
if ischar(fn)
    fn = {fn};
elseif iscell(fn)
    if any(cellfun(@ischar, fn)==0)
        error('Variable ''fn'' must be a string or cell array of strings.');
    end
else
    error('Variable ''fn'' must be a string or cell array of strings.');
end

%% Merge results from multiple files into one
for k=1:numel(fn)
    if k==1 % Initialize results structure
        results = load([path2DB fn{k}]);
        fn_results = fieldnames(results);
        fn_results(strcmp(fn_results,'idxAnalyze') | strcmp(fn_results,'params')) = [];
        fn_params = fieldnames(results.params);
    else % Concatenate new results to old results
        results2 = load([path2DB fn{k}]);
        for f=1:numel(fn_results)
            results.(fn_results{f}) = [results.(fn_results{f});results2.(fn_results{f})];
        end
        for f=1:numel(fn_params)
            results.params.(fn_params{f}) = [results.params.(fn_params{f})(:);results2.params.(fn_params{f})(:)];
        end
        
    end
end

clear results2 fn fn2 k f;

% % Optionally save these results to a single file
% save([path2DB 'SummaryResults.mat'],'-struct','results');


%%  --------------- EFFICIENCY CURVE ---------------
%=== Sub-sample and plot the efficiency curve

% Set number of iterations in each of our subsamples searches
% Note: if the total number of searches is not evenly divisible by numK,
% then up to up to (numK-1) of the random searches are ignored
numK = 2.^(1:7); K = numel(numK);

f1_data_db1 = nan(numK(end)/numK(1),K);
f1_data_db2 = nan(numK(end)/numK(1),K);
se_data_db1 = nan(numK(end)/numK(1),K);
se_data_db2 = nan(numK(end)/numK(1),K);
ppv_data_db1 = nan(numK(end)/numK(1),K);
ppv_data_db2 = nan(numK(end)/numK(1),K);

% Pre-allocate index of best performing algorithm for each repetition
idxBest = nan(numK(end)/numK(1),K);
%rng(1934,'twister'); % fix the RNG seed for repeatable results

f1_db1 = results.F1Score;
f1_db2 = results.F1Score_DB2;
se_db1 = results.Se;
se_db2 = results.Se_DB2;
ppv_db1 = results.PPV;
ppv_db2 = results.PPV_DB2;
N = size(f1_db1,1);


%=== Subsample the overall random search and calculate the performance of
%the subsample's best hyperparameters as measured by the F1 measure on DB1 
for k=1:K
    %=== Calculate the maximum number of folds, and sub-sample data to this
    N_PER_FOLD = floor(N / numK(k));
    N_TOTAL = N_PER_FOLD * numK(k); % number of total observations used
    F = N_TOTAL / numK(k); % number of folds
    
    %=== Create equal number of indices for each fold
    [~,idxFold] = sort(rand(N_TOTAL,1));
    idxFold = mod(idxFold, F)+1;
    
    %=== Create matching indices for original data
    [~,idxK] = sort(rand(N,1));
    idxK = idxK(1:N_TOTAL);
    
    %=== For each "fold", find the maximum performing F1 on DB1
    for f=1:F
        idxF = idxK(idxFold==f);
        [f1_data_db1(f,k),idxBest(f,k)] = max(mean(f1_db1(idxF,:),2));
        idxBest(f,k) = idxF(idxBest(f,k)); % convert back to original idx
        f1_data_db2(f,k) = mean(f1_db2(idxBest(f,k),:),2);
        
        se_data_db1(f,k) = mean(se_db1(idxBest(f,k),:),2);
        se_data_db2(f,k) = mean(se_db2(idxBest(f,k),:),2);
        ppv_data_db1(f,k) = mean(ppv_db1(idxBest(f,k),:),2);
        ppv_data_db2(f,k) = mean(ppv_db2(idxBest(f,k),:),2);
    end
    
end
f1_data_db1 = f1_data_db1*100;
f1_data_db2 = f1_data_db2*100;
se_data_db1 = se_data_db1*100;
se_data_db2 = se_data_db2*100;
ppv_data_db1 = ppv_data_db1*100;
ppv_data_db2 = ppv_data_db2*100;

%% Print out results @ a certain number of iterations
k = 5;
N_PER_FOLD = floor(N / numK(k));
N_TOTAL = N_PER_FOLD * numK(k); % number of total observations used
F = N_TOTAL / numK(k); % number of folds
idxReport = median(f1_data_db1(1:F,k));
[~,idxReport] = min(abs(f1_data_db1(1:F,k) - idxReport));
idxReport = idxBest(idxReport,k);

fprintf('Performance of ESN at %d iterations\n', numK(k));
fprintf('\tSe\t\tPPV\t\tF1\n\n');
fprintf('DB1\t');%DB1
fprintf('%2.1f %1.2f\t%2.1f %1.2f\t%2.1f %1.2f\n', ...
    results.Se(idxReport)*100,std(se_data_db1(1:F,k)),...
    results.PPV(idxReport)*100,std(ppv_data_db1(1:F,k)),...
    results.F1Score(idxReport)*100,std(f1_data_db1(1:F,k)));
    
fprintf('DB2\t');%DB2

fprintf('%2.1f %1.2f\t%2.1f %1.2f\t%2.1f %1.2f\n', ...
    results.Se_DB2(idxReport)*100,std(se_data_db2(1:F,k)),...
    results.PPV_DB2(idxReport)*100,std(ppv_data_db2(1:F,k)),...
    results.F1Score_DB2(idxReport)*100,std(f1_data_db2(1:F,k)));

fprintf('\n');
fprintf('Parameters:\n');
fn_param = fieldnames(results.params);
for k=1:numel(fn_param)
    if isnumeric(results.params.(fn_param{k}))
    fprintf('%8g\t%s\n',results.params.(fn_param{k})(idxReport),fn_param{k})
    else
    fprintf('%8s\t%s\n',results.params.(fn_param{k}){idxReport},fn_param{k})
    end
end
%% Plot the efficiency curve
figure(1); clf; hold all;
hax = gca;
for k=1:K
    N_PER_FOLD = floor(N / numK(k));
    N_TOTAL = N_PER_FOLD * numK(k); % number of total observations used
    F = N_TOTAL / numK(k); % number of folds
    curr_xi = ones(F,1)*numK(k);
    curr_xi = log(curr_xi); % log-scale
    
    if F > 10 % more than 10 points to plot
        % do box plot
        h=boxplot(f1_data_db1(1:F,k),'positions',curr_xi,...
            'Colors','b','symbol','bx','labels',{''});
        set(h,'HandleVisibility','off');
        h=boxplot(f1_data_db2(1:F,k),'positions',curr_xi,...
            'Colors','r','symbol','r+','labels',{''});
        set(h,'HandleVisibility','off');
    else % plot the dots
        plot(curr_xi,f1_data_db1(1:F,k),'bx','MarkerSize',6,'HandleVisibility','off');
        plot(curr_xi,f1_data_db2(1:F,k),'r+','MarkerSize',6,'HandleVisibility','off');
    end
end
set(gca,'XTick',log(numK),'XTickLabel',arrayfun(@num2str, numK, 'UniformOutput', false));
ylbl = ylabel('F measure'); set(ylbl,'Units','Normalized');
xlbl = xlabel('');

% From script ESN_ReportGridSearchResults we have the best grid results
% 	DB1                     DB2
% Se	PPV     F1          Se      PPV     F1		File name
% 97.0	97.2	97.1		88.8	86.2	87.5	GridResult_221113_0123362.mat

plot([0,log(numK(end))+0.5],[97.1,97.1],'--','Color',[0 0 0.8], 'LineWidth',2);
plot([0,log(numK(end))+0.5],[87.9,87.9],'-.','Color',[0.8 0 0], 'LineWidth',2);
set(gca,'XLim',[log(numK(1)-.2),log(numK(end)+1)],'YLim',[60 100]);
hLegend=legend('DB1','DB2','Location','SouthEast');

%% Pretty up the figure
set(hLegend,'FontName','AvantGarde','FontSize',18,'TextColor',[.3 .3 .3],'FontWeight','bold',...
    'Box','off');

set(hax, ...
    'FontName'   , 'AvantGarde', ...
    'FontSize'    , 14          );

set(hax             , ...
    'Box'         , 'off'       , ...
    'TickDir'     , 'out'       , ...
    'XColor'      , [.3 .3 .3], ...
    'YColor'      , [.3 .3 .3]);

set([get(hax,'YLabel'), get(hax,'XLabel')]  , ...
    'FontSize'   , 16);

xpos = get(xlbl,'Position');
xpos(2) = xpos(2)-10;
set(xlbl,'Position',xpos);

% %=== Optionally save to file
% saveas(1,[path2DB 'EfficiencyCurve.fig']);
