%ESN_INTRINSICDIMENSIONALITY    Analyzes the effect of hyperparameters on
% a performance metric using a Gaussian Process.
%
% This code requires the gpml toolbox. Run this code as such will run it 
% on the example of the ESN hyperparameters as per the work presented in
% [1].
%
% References
% [1] Behar, Joachim and Alistair Johnson, Julien Oster, Gari D. Clifford.
% An Echo State Neural Network for Foetal Electrocardiogram Extraction 
% Optimised by Random Search. NIPS Lake Tahoe, Nevada, United States, 5-10
% December 2013. 
% [2] Bergstra, James and Yoshua Bengio. "Random search for hyper-parameter
% optimization." The Journal of Machine Learning Research 13 (2012): 281-305.
%
% The following variables are set by the user:
% path2DB   - Path containing the ESN data/results
% fn        - String (or cell array of strings) containing the name of
%           result files. The result files should contain at a minimum:
%               F1Score         - Nx1 vector of DB1 F1 measures
%               F1Score_DB2     - Nx1 vector of DB2 F1 measures
%               params          - Structure containing Nx1 vectors of each
%                               hyperparameter that changed during search
%       ... where N is the number of random searches performed

%	Copyright 2013 Alistair Johnson

%	$LastChangedBy: Joachim Behar $
%	$LastChangedDate: 2013-12-06 $
%	$Revision: 961 $
%	Originally written on PCWIN64 by Alistair Johnson, 11-Nov-2013 23:31:16
%	Contact: alistairewj@gmail.com

%%  --------------- HEADER ---------------
clearvars -except path2DB fn; close all; clc;

addpath(genpath('gpml-matlab/'))

if exist('fn','var')~=1 % Default file name
fn = 'SummaryResults.mat'; % repetition of 512 runs
end

if exist('path2DB','var')~=1 % Default path
path2DB = ''; % absolute path to database
end


% Set the paths containing the ESN database
DOWN_SAMPLE = 1;                 % run all the algo at 250Hz or 1000Hz
N_DB1 = 14; % records in database 1
N_DB2 = 11; % records in database 2

addpath(genpath(path2DB)); 

%%  --------------- LOAD RESULTS ---------------  
%=== Ensure the file name(s) are valid
if ischar(fn)
    fn = {fn};
elseif iscell(fn)
    if any(cellfun(@ischar, fn)==0)
        error('Variable ''fn'' must be a string or cell array of strings.');
    end
else
    error('Variable ''fn'' must be a string or cell array of strings.');
end

%% Merge results from multiple files into one
for k=1:numel(fn)
    if k==1 % Initialize results structure
        results = load([path2DB fn{k}]);
        fn_results = fieldnames(results);
        fn_results(strcmp(fn_results,'idxAnalyze') | strcmp(fn_results,'params')) = [];
        fn_params = fieldnames(results.params);
    else % Concatenate new results to old results
        results2 = load([path2DB fn{k}]);
        for f=1:numel(fn_results)
            results.(fn_results{f}) = [results.(fn_results{f});results2.(fn_results{f})];
        end
        for f=1:numel(fn_params)
            results.params.(fn_params{f}) = [results.params.(fn_params{f})(:);results2.params.(fn_params{f})(:)];
        end
        
    end
end

clear results2 fn fn2 k f;

% % Optionally save these results to a single file
% save([path2DB 'SummaryResults.mat'],'-struct','results');

%% Extract from the results structure variables necessary for this function
f1_db1 = results.F1Score;
f1_db2 = results.F1Score_DB2;
hypHeader = fieldnames(results.params);
D = numel(hypHeader);
N = size(f1_db1,1);

%% Calculate instrinsic dimensionality using a Gaussian Process

%=== Load hyperparameters into a matrix
idxRemove = false(1,D);
hyperparams = nan(N,D);
for d=1:D
    if numel(unique(results.params.(hypHeader{d})))<=1
        %=== This hyperparameter was not searched
        fprintf('Not analyzing %s, which kept a constant value for all searches.\n',hypHeader{d});
        idxRemove(d) = true;
    else
    hyperparams(:,d) = results.params.(hypHeader{d});
    end
end

hypHeader(idxRemove) = [];
hyperparams(:,idxRemove) = [];

% ESN specific check - nInternalUnits should be analyzed logarithmically
hypFlag = ismember(hypHeader,{'nInternalUnits'});
hypRange = [min(hyperparams,[],1);max(hyperparams,[],1)];

% DB1!
figure(2); clf;
[ gp_results1 ] = IntrinsicDimensionality(f1_db1, hyperparams, hypRange, hypFlag, hypHeader);

% DB2!
figure(3); clf;
[ gp_results2 ] = IntrinsicDimensionality(f1_db2, hyperparams, hypRange, hypFlag, hypHeader);

%% Replot both DBs on one figure with pretty labels
figure(4); close(4); figure(4);
set(gcf,'OuterPosition',[395         440        1005         607]);

%=== For purely asthetic reasons, we replace some words with symbols
hypHeaderMap = {'seeds','$s$';
    'leakage','$\alpha$';
    'spectralRadius','$\rho$';
    'inputScaling','$\gamma$';
    'fbas','$f_{b}$';
    'fhigh','$f_{h}$';
    'nInternalUnits','$M$'};
[idxExist,idxMap] = ismember(hypHeader,hypHeaderMap(:,1));
hypHeaderPretty = hypHeader;
hypHeaderPretty(idxExist) = hypHeaderMap(idxMap(idxExist),2);

% set at least 8 colours and repeat them as needed
col = [1,0,0; % red
    0,0,1; % blue
    0,0.8,0; % green
    0.4,0.4,0; % dark yellow
    0,0.8,0.8; % cyan
    0.4,0,0.8; % dark magenta
    0.8,0.4,1; % light magenta
    0.4,0.4,1]; % lilac

%% First subplot: DB1
h1=subplot(1,2,1);  hold all;
boxplot(1./gp_results1.ell,'colors',col,'orientation','vertical','symbol','+');
ylabel('relevance ( 1 / length scale )','FontSize',18);
% Labels on plot have been input!
xTicks = 1:numel(hypHeader);
yTicks = 0:0.5:4;
set(gca,'XTick',xTicks,'XTickLabel',{''},'YTick',yTicks,'YLim',[-0.1,4],'FontSize',14,...
    'GridLineStyle',':','YGrid','on');

%% Set tick labels using text so it can parse LaTeX formats
% Reset the xtick labels in desired font 
minY = min(yTicks);

% these offsets are used to place the latex variable names
verticalOffset = 0.25;
hypHeaderOffset = [0.1,0.12,0.1,0.11,0.20,0.20,0.23];

for xx = 1:length(xTicks)
     text(xTicks(xx) + hypHeaderOffset(xx), minY - verticalOffset, hypHeaderPretty{xx},...
         'HorizontalAlignment','Right','interpreter', 'latex',...
         'FontName'   , 'AvantGarde', 'FontSize',14,...
         'Color'      , [.3 .3 .3]);  
end


%% Second subplot: DB2
h2=subplot(1,2,2);  hold all;
boxplot(1./gp_results2.ell,'colors',col,'orientation','vertical','symbol','+');
% ylabel('relevance ( 1 / length scale )','FontSize',18);

xTicks = 1:numel(hypHeader);
yTicks = 0:0.5:4;
set(gca,'XTick',xTicks,'XTickLabel',{''},'YTick',yTicks,'YLim',[-0.1,4],'FontSize',14,...
    'GridLineStyle',':','YGrid','on');

minY = min(yTicks);
for xx = 1:length(xTicks)
     text(xTicks(xx) + hypHeaderOffset(xx), minY - verticalOffset, hypHeaderPretty{xx},...
         'HorizontalAlignment','Right','interpreter', 'latex',...
         'FontName'   , 'AvantGarde', 'FontSize',14,...
         'Color'      , [.3 .3 .3]);  
end

%% Pretty up the figure
set([h1,h2], ...
    'FontName'   , 'AvantGarde');

set([h1,h2]             , ...
    'FontSize'   , 16           );
set([get(h1,'YLabel'), get(h2,'YLabel')]  , ...
    'FontSize'   , 14  ,...
    'FontWeight', 'bold');

%=== set some defaults on the tick length
set([h1,h2] , ...
    'Box'         , 'off'     , ...
    'TickDir'     , 'out'     , ...
    'XColor'      , [.3 .3 .3], ...
    'YColor'      , [.3 .3 .3]);


%% Optionally save the figure (uncomment if needed)
% saveas(4,[path2DB 'IntrinsicDimensionality.fig']);