function [ results ] = IntrinsicDimensionality(perf, hyperparams, hypRange, hypFlag, hypHeader)
%INTRINSICDIMENSIONALITY	Estimate the intrinsic dimensionality of
%various hyperparameters by measuring their relevance to performance using
%a Gaussian Process.
%	[ RESULTS ] = IntrinsicDimensionality(PERF, HYPERPARAMS) measures the
%	relevance of each column of matrix HYPERPARAMS to column vector of
%	performances PERF. The hyperparameters are scaled to the [0,1] interval
%	to ensure they are comparable. RESULTS contains three fields pertaining
%	to the GPs trained. Since there exist many local minima, GP development
%	is repeated on subsets of the hyperparameters.
%	
%	[ RESULTS ] = IntrinsicDimensionality(PERF, HYPERPARAMS, HYPRANGE,
%	HYPFLAG) allows specification of the range of hyperparameters
%	(HYPRANGE) for scaling to the [0,1] interval. HYPFLAG specifies whether
%	the parameters should be log-transformed before analysis, as some
%	hyperparameters may have been searched exponentially.
%	
%	[ RESULTS ] = IntrinsicDimensionality(PERF, HYPERPARAMS, HYPRANGE,
%	HYPFLAG, HYPHEADER) allows inputting labels for each column of the
%	hyperparameter. These labels are used for plotting purposes.
%
%	Inputs:
%		PERF			- Nx1 vector of performances
%		HYPERPARAMS		- NxD matrix, with D columns of N hyperparameters
%		HYPRANGE		- 2xD matrix of minimum (first row) and maximum
%						(second row) possible hyperparameters
%		HYPFLAG			- 1xD vector of logical flags, indicating the
%						hyperparameter should be log-transformed
%		HYPHEADER		- 1xD cell array of strings
%
%	Outputs:
%		RESULTS			- Structure containing three fields:
%			gp			Mx1 cell of trained GPs using subsets of data
%			ard			MxD matrix of ARD measurements for each hyperparam
%			ll			Mx1 vector of log-likelihoods from the GPs
%
%		Note M is the number of GP training repetitions.
%
%	Example
% 		load('C:\Users\Alistair\Documents\MATLAB\PredictionAlgorithm\RandomParamsWithDBPerformance.mat')
% 		hyperparams = [fbas,fhigh,feedbackScaling,inputScaling,leakage,nInternalUnits,spectralRadius,seeds];
% 		perf = F1Score;
% 		hyperparams_ranges = [1,50;50,120;0,1;0,1;0,1;10,150;0,1;0,10000];
% 		hyperparams_flag = false(1,8);
% 		hypHeader = {'fbas','fhigh','feedback scaling','input scaling',...
% 			'leakage','neurons','spectral radius','seeds'};
% 		[ results ] = IntrinsicDimensionality(perf, hyperparams, ...
% 			hyperparams_ranges, hyperparams_flag, hypHeader);
% 	
%	See also MINIMIZE GP

%	References:
%		Gaussian Processes for Machine Learning. 
%		Rasmussen and Williams, 2006.
%
%		Random search for Hyperparameter Optimization. 
%		Bergstra and Bengio, JMLR 2012.

%	Copyright 2013 Alistair Johnson

%	$LastChangedBy: alistair $
%	$LastChangedDate: 2013-11-28 16:58:23 +0000 (Thu, 28 Nov 2013) $
%	$Revision: 961 $
%	Originally written on PCWIN64 by Alistair Johnson, 11-Nov-2013 23:02:44
%	Contact: alistairewj@gmail.com

%=== GP parameters that are user configurable
N_REPEAT = 50; % number of GP repetitions
N_SUBSAMPLE = 0.8; % percentage of data to subsample for each GP repetition
N_ITERATIONS = 100; % Number of iterations in the GP linesearch

[N,D] = size(hyperparams);
M = size(perf,1);
results = struct();
if N~=M
    error('Number of performance measurements must equal number of hyperparameters');
end


if nargin<3 || isempty(hypRange)
    hypRange = [min(hyperparams,[],1)',max(hyperparams,[],1)']';
else
    if size(hypRange,1) == D && size(hypRange,2) == 2
        hypRange = hypRange';
    elseif size(hypRange,2) == D && size(hypRange,1) == 2
        % right dimensionality
    else
        error('Hyperparameter ranges for scaling must be a 2xD or Dx2 matrix where D is the number of hyperparameters');
    end
end

%=== some parameters need to be transformed to the log scale that it is
%searched on
if nargin<4 || isempty(hypFlag)
    hypFlag = false(1,D);
end
hyperparams(:,hypFlag) = log(hyperparams(:,hypFlag));
hypRange(:,hypFlag) = log(hypRange(:,hypFlag));


hyperparams = bsxfun(@rdivide, bsxfun(@minus, hyperparams, hypRange(1,:)), hypRange(2,:));
if any(any(hyperparams<0)) || any(any(hyperparams>1))
    error('Hyperparameters not correctly scaled to [0,1] interval. Check range values');
end

%=== Initialize GP parameters
covFcn = @covSEard;
likFcn = @likGauss;
infFcn = @infExact;

%=== Initialize sub-sampling parameters
% Many local minima exist - thus we repeat the development with subsets of
% data
N_SUBSAMPLE = round(N_SUBSAMPLE*N);

gp_trained = cell(N_REPEAT,1);
loglik = nan(N_REPEAT,1);

for k = 1:N_REPEAT
    
    %=== Subsample data
    [~,idxCurr] = sort(rand(N,1),1);
    idxCurr = idxCurr(1:N_SUBSAMPLE);
    curr_perf = perf(idxCurr,:);
    curr_hyp = hyperparams(idxCurr,:);
       
    %=== Randomly set initial length scales for each parameter
    l_init = rand(1,D)*1.9 + 0.1;
    l_init = log(l_init); % hyperparameter expected in log(ll) format
    
    %=== Initialize hyperparameter structure for GP code
    hyp = struct();
    hyp.cov = [l_init(:);7.3891];
    hyp.lik = 1;        

    gp_trained{k} = minimize(hyp, @gp, N_ITERATIONS, infFcn,[], covFcn, likFcn, ...
        curr_hyp, curr_perf);
    loglik(k) = gp(gp_trained{k}, infFcn, [], covFcn, likFcn, curr_hyp, curr_perf);
end

results.gp = gp_trained;
results.loglik = loglik;

%=== Plot final results by merging cov estimates and box-plotting
ell = cellfun(@(x) x.cov(1:end-1)', results.gp, 'UniformOutput',false);
ell = vertcat(ell{:});
ell = exp(ell);
results.ell = ell;

% set at least 8 colours and repeat them as needed
col = [1,0,0; % red
    0,0,1; % blue
    0,0.8,0; % green
    0.4,0.4,0; % dark yellow
    0,0.8,0.8; % cyan
    0.4,0,0.8; % dark magenta
    0.8,0.4,1; % light magenta
    0.4,0.4,1]; % lilac
if size(col,1)<D
    % we don't have enough colours - need to replicate the matrix
    col = repmat(col, ceil(D/size(col,1)), 1);
end

boxplot(1./ell,'colors',col,'orientation','horizontal','symbol','+');

if nargin>4
    % Labels on plot have been input!
    set(gca,'YTick',1:D,'YTickLabel',hypHeader);
end

set(gca,'FontSize',14,'FontName','AvantGarde');
xlabel('relevance ( 1 / length scale )','FontSize',14,'FontName','AvantGarde');
end